/*
 * Copyright (c) 2003, the JUNG Project and the Regents of the University of
 * California All rights reserved.
 *
 * This software is open-source under the BSD license; see either "license.txt"
 * or http://jung.sourceforge.net/license.txt for a description.
 */
package edu.uci.ics.jung.algorithms.layout;

import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.ConcurrentModificationException;
import java.util.Map;

import com.google.common.base.Function;
import com.google.common.collect.MapMaker;

import edu.uci.ics.jung.algorithms.layout.util.RandomLocationTransformer;
import edu.uci.ics.jung.algorithms.util.IterativeContext;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.Pair;

/**
 * Implements the Fruchterman-Reingold force-directed algorithm for node layout.
 * 
 * <p>Behavior is determined by the following settable parameters:
 * <ul>
 * <li/>attraction multiplier: how much edges try to keep their vertices together
 * <li/>repulsion multiplier: how much vertices try to push each other apart
 * <li/>maximum iterations: how many iterations this algorithm will use before stopping
 * </ul>
 * Each of the first two defaults to 0.75; the maximum number of iterations defaults to 700.
 *
 * <p>
 * <br>
 * EDIT Johannes Zink am 14.05.2015 (und kleine Teile zeitlich danach):
 * Ich will diese Algorithmus, so anpassen, dass ich die Berechnung der Abstossung
 * abaendere. Daher erstelle ich eine Klasse, die von dieser Klasse erbt.
 * Ich habe aber auf viele Variablen/Methoden in dieser Klasse sonst keinen Zugriff.
 * Deshalb habe ich folgende Aenderungen vorgenommen:
 * <br>- "private int currentIteration" zu "protected int currentIteration" (Instanzvariable)
 * <br>- "private double EPSILON" zu "protected double EPSILON" (Instanzvariable)
 * <br>- "private double repulsion_constant" zu "protected double repulsion_constant" (Instanzvariable)
 * <br>- "private void cool()" zu "protected void cool()" (Instanzmethode)
 * <br>- In der Klasse"FRLayout.FRVertexData": "protected void offset(double x, double y)" zu "public void offset(double x, double y)"
 * (Instanzmethode in der hier definerten, neuen Klasse FRVertexData)
 * <br>- "private Map<V, FRVertexData> frVertexData" zu "protected Map<V, FRVertexData> frVertexData" (Instanzvariable)
 * <p>
 * EDIT Johannes Zink am 24.06.2015:
 * <br>- "private double max_dimension" zu "protected double max_dimension" (Instanzvariable)
 * <br>- die Klasse "FRVertexData" von protected zu public
 * <br>
 * <p>
 * EDIT Johannes Zink am 20.07.2015:
 * <br>- "private double temperature" zu "protected double temperature" (Instanzvariable)
 * <br>- "private double forceConstant" zu "protected double forceConstant" (Instanzvariable)
 * <br>- "private double attraction_constant" zu "protected double attraction_constant" (Instanzvariable)
 * <br>- "private double attraction_multiplier" zu "protected double attraction_multiplier" (Instanzvariable)
 * <br>- "private double repulsion_multiplier" zu "protected double repulsion_multiplier" (Instanzvariable)
 * <br>- "protected double FRVertexData.norm()" zu "public double FRVertexData.norm()" (Methode der Klasse FRVertexData)
 * <br>
 * <p>
 * EDIT Johannes Zink am 26.07.2015:
 * <br>- "private int mMaxIterations" zu "protected int mMaxIterations" (Instanzvariable)
 * <p>
 *
 * @see "Fruchterman and Reingold, 'Graph Drawing by Force-directed Placement'"
 * @see "http://i11www.ilkd.uni-karlsruhe.de/teaching/SS_04/visualisierung/papers/fruchterman91graph.pdf"
 * @author Scott White, Yan-Biao Boey, Danyel Fisher
 */
public class FRLayout<V, E> extends AbstractLayout<V, E> implements IterativeContext {

    protected double forceConstant;

    protected double temperature;

    protected int currentIteration;

    protected int mMaxIterations = 700;

    protected Map<V, FRVertexData> frVertexData =
    	new MapMaker().makeComputingMap(new Function<V,FRVertexData>(){
//			@Override
			public FRVertexData apply(V arg0) {
				return new FRVertexData();
			}});
//    	LazyMap.decorate(new HashMap<V,FRVertexData>(), new Supplier<FRVertexData>() {
//    		public FRVertexData create() {
//    			return new FRVertexData();
//    		}});

    protected double attraction_multiplier = 0.50;

    protected double attraction_constant;

    protected double repulsion_multiplier = 0.99;

    protected double repulsion_constant;

    protected double max_dimension;

    /**
     * Creates an instance for the specified graph.
     */
    public FRLayout(Graph<V, E> g) {
        super(g);
    }

    /**
     * Creates an instance of size {@code d} for the specified graph.
     */
    public FRLayout(Graph<V, E> g, Dimension d) {
        super(g, new RandomLocationTransformer<V>(d), d);
        initialize();
        max_dimension = Math.max(d.height, d.width);
    }

	@Override
	public void setSize(Dimension size) {
		if(initialized == false) {
			setInitializer(new RandomLocationTransformer<V>(size));
		}
		super.setSize(size);
        max_dimension = Math.max(size.height, size.width);
	}

	/**
	 * Sets the attraction multiplier.
	 */
	public void setAttractionMultiplier(double attraction) {
        this.attraction_multiplier = attraction;
    }

	/**
	 * Sets the repulsion multiplier.
	 */
    public void setRepulsionMultiplier(double repulsion) {
        this.repulsion_multiplier = repulsion;
    }

	public void reset() {
		doInit();
	}

    public void initialize() {
    	doInit();
    }

    private void doInit() {
    	Graph<V,E> graph = getGraph();
    	Dimension d = getSize();
    	if(graph != null && d != null) {
    		currentIteration = 0;
    		temperature = d.getWidth() / 10;

    		forceConstant =
    			Math
    			.sqrt(d.getHeight()
    					* d.getWidth()
    					/ graph.getVertexCount());

    		attraction_constant = attraction_multiplier * forceConstant;
    		repulsion_constant = repulsion_multiplier * forceConstant;
    	}
    }

    protected double EPSILON = 0.000001D;

    /**
     * Moves the iteration forward one notch, calculation attraction and
     * repulsion between vertices and edges and cooling the temperature.
     */
    public synchronized void step() {
        currentIteration++;

        /**
         * Calculate repulsion
         */
        while(true) {

            try {
                for(V v1 : getGraph().getVertices()) {
                    calcRepulsion(v1);
                }
                break;
            } catch(ConcurrentModificationException cme) {}
        }

        /**
         * Calculate attraction
         */
        while(true) {
            try {
                for(E e : getGraph().getEdges()) {

                    calcAttraction(e);
                }
                break;
            } catch(ConcurrentModificationException cme) {}
        }


        while(true) {
            try {
                for(V v : getGraph().getVertices()) {
                    if (isLocked(v)) continue;
                    calcPositions(v);
                }
                break;
            } catch(ConcurrentModificationException cme) {}
        }
        cool();
    }

    protected synchronized void calcPositions(V v) {
        FRVertexData fvd = getFRData(v);
        if(fvd == null) return;
        Point2D xyd = apply(v);
        double deltaLength = Math.max(EPSILON, fvd.norm());

        double newXDisp = fvd.getX() / deltaLength
                * Math.min(deltaLength, temperature);

        if (Double.isNaN(newXDisp)) {
        	throw new IllegalArgumentException(
                "Unexpected mathematical result in FRLayout:calcPositions [xdisp]"); }

        double newYDisp = fvd.getY() / deltaLength
                * Math.min(deltaLength, temperature);
        xyd.setLocation(xyd.getX()+newXDisp, xyd.getY()+newYDisp);

        double borderWidth = getSize().getWidth() / 50.0;
        double newXPos = xyd.getX();
        if (newXPos < borderWidth) {
            newXPos = borderWidth + Math.random() * borderWidth * 2.0;
        } else if (newXPos > (getSize().getWidth() - borderWidth)) {
            newXPos = getSize().getWidth() - borderWidth - Math.random()
                    * borderWidth * 2.0;
        }

        double newYPos = xyd.getY();
        if (newYPos < borderWidth) {
            newYPos = borderWidth + Math.random() * borderWidth * 2.0;
        } else if (newYPos > (getSize().getHeight() - borderWidth)) {
            newYPos = getSize().getHeight() - borderWidth
                    - Math.random() * borderWidth * 2.0;
        }

        xyd.setLocation(newXPos, newYPos);
    }

    protected void calcAttraction(E e) {
    	Pair<V> endpoints = getGraph().getEndpoints(e);
        V v1 = endpoints.getFirst();
        V v2 = endpoints.getSecond();
        boolean v1_locked = isLocked(v1);
        boolean v2_locked = isLocked(v2);

        if(v1_locked && v2_locked) {
        	// both locked, do nothing
        	return;
        }
        Point2D p1 = apply(v1);
        Point2D p2 = apply(v2);
        if(p1 == null || p2 == null) return;
        double xDelta = p1.getX() - p2.getX();
        double yDelta = p1.getY() - p2.getY();

        double deltaLength = Math.max(EPSILON, Math.sqrt((xDelta * xDelta)
                + (yDelta * yDelta)));

        double force = (deltaLength * deltaLength) / attraction_constant;

        if (Double.isNaN(force)) { throw new IllegalArgumentException(
                "Unexpected mathematical result in FRLayout:calcPositions [force]"); }

        double dx = (xDelta / deltaLength) * force;
        double dy = (yDelta / deltaLength) * force;
        if(v1_locked == false) {
        	FRVertexData fvd1 = getFRData(v1);
        	fvd1.offset(-dx, -dy);
        }
        if(v2_locked == false) {
        	FRVertexData fvd2 = getFRData(v2);
        	fvd2.offset(dx, dy);
        }
    }

    protected void calcRepulsion(V v1) {
        FRVertexData fvd1 = getFRData(v1);
        if(fvd1 == null)
            return;
        fvd1.setLocation(0, 0);

        try {
            for(V v2 : getGraph().getVertices()) {

//                if (isLocked(v2)) continue;
                if (v1 != v2) {
                    Point2D p1 = apply(v1);
                    Point2D p2 = apply(v2);
                    if(p1 == null || p2 == null) continue;
                    double xDelta = p1.getX() - p2.getX();
                    double yDelta = p1.getY() - p2.getY();

                    double deltaLength = Math.max(EPSILON, Math
                            .sqrt((xDelta * xDelta) + (yDelta * yDelta)));

                    double force = (repulsion_constant * repulsion_constant) / deltaLength;

                    if (Double.isNaN(force)) { throw new RuntimeException(
                    "Unexpected mathematical result in FRLayout:calcPositions [repulsion]"); }

                    fvd1.offset((xDelta / deltaLength) * force,
                            (yDelta / deltaLength) * force);
                }
            }
        } catch(ConcurrentModificationException cme) {
            calcRepulsion(v1);
        }
    }

    protected void cool() {
        temperature *= (1.0 - currentIteration / (double) mMaxIterations);
    }

    /**
     * Sets the maximum number of iterations.
     */
    public void setMaxIterations(int maxIterations) {
        mMaxIterations = maxIterations;
    }

    protected FRVertexData getFRData(V v) {
        return frVertexData.get(v);
    }

    /**
     * This one is an incremental visualization.
     */
    public boolean isIncremental() {
        return true;
    }

    /**
     * Returns true once the current iteration has passed the maximum count,
     * <tt>MAX_ITERATIONS</tt>.
     */
    public boolean done() {
        if (currentIteration > mMaxIterations || temperature < 1.0/max_dimension)
        {
            return true;
        }
        return false;
    }

    public static class FRVertexData extends Point2D.Double
    {
        public void offset(double x, double y)
        {
            this.x += x;
            this.y += y;
        }

        public double norm()
        {
            return Math.sqrt(x*x + y*y);
        }
     }
}